@@ -6,7 +6,7 @@ using namespace arma;
 using namespace Rcpp;
 
 // define integers for function names
-enum array_act { a_sum, a_prod, a_all, a_any, a_min, a_max, a_mean, a_median, a_sd, a_var, a_norm, a_trapz, a_normalise, a_cumsum, a_cumprod, a_multv, a_divv, a_addv, a_subv, a_diff, a_range };
+enum array_act { a_sum, a_prod, a_all, a_any, a_min, a_max, a_mean, a_median, a_sd, a_var, a_norm, a_trapz, a_normalise, a_cumsum, a_cumprod, a_multv, a_divv, a_addv, a_subv, a_diff, a_range, a_conv, a_quantile };
 
 // result of a call on a vector is a real scalar
 #define make_call(act) \
@@ -47,28 +47,57 @@ case a_##act: \
 //' High Performance Variant of apply()
 //'
 //'  High performance variant of apply() for a fixed set of functions.
-//'  Considerable speedup is a trade-off for universality, user defined
-//'  functions cannot be used with arrApply. However, 20 most currently employed
-//'  functions are available for usage. They can be divided in three types:
-//'  reducing functions (like mean(), sum() etc., giving a scalar when applied to a vector),
-//'  mapping function (like normalise(), cumsum() etc., giving a vector of the same length
-//'  as the input vector) and finally, vector reducing function (like diff() which produces
-//'  result vector of a length different from the length of input vector).
+//'  Considerable speedup obtained by this implementation is a trade-off for universality, user defined
+//'  functions cannot be used with arrApply. However, about 20 most currently employed
+//'  functions are available for usage. They can be divided in three types: \itemize{
+//'    \item reducing functions (like mean(), sum() etc., giving a scalar when applied to a vector);
+//'    \item mapping function (like normalise(), cumsum() etc., giving a vector of the same length
+//'        as the input vector)
+//'    \item and finally, vector reducing function (like diff() which produces
+//'        result vector of a length different from the length of input vector).
+//'  }
 //'  Optional or mandatory additional arguments required by some functions
 //'  (e.g. norm type for norm() or normalise() functions) can be
 //'  passed as named arguments in '...'.
 //' 
 //' The following functions can be used as argument 'fun' (brackets
-//' [] indicate additional parameters that can be passed in '...'):
-//'  - reducing functions: sum(), prod(), all(), any(), min(), max(),
-//'    mean(), median(), sd() [norm_type], var() [norm_type], norm() [p],
-//'    trapz() [x] (trapezoidal integration with respect to spacing in x,
-//'    if x is provided, otherwise unit spacing is used), range();
-//'  - mapping functions: normalise() [p], cumsum(), cumprod(), multv() [v]
-//'    (multiply a given dimension by a vector v, term by term), divv() [v]
-//'    (divide by a vector v), addv() [v] (add a vector v), subv() [v] (subtract
-//'    a vector v);
-//'  - vector reducing function: diff() [k].
+//' [] indicate additional parameters that can be passed in '...'): \itemize{
+//'     \item reducing functions: \itemize{
+//'         \item sum()
+//'         \item prod()
+//'         \item all()
+//'         \item any()
+//'         \item min()
+//'         \item max()
+//'         \item mean()
+//'         \item median()
+//'         \item sd() [norm_type]
+//'         \item var() [norm_type]
+//'         \item norm() [p],
+//'         \item trapz() [x] (trapezoidal integration with respect to spacing in x,
+//'            if x is provided, otherwise unit spacing is used)
+//'         \item range();
+//'       }
+//'     \item mapping functions:\itemize{
+//'         \item normalise() [p]
+//'         \item cumsum()
+//'         \item cumprod()
+//'         \item multv() [v]
+//'            (multiply a given dimension by a vector v, term by term)
+//'         \item divv() [v]
+//'            (divide by a vector v)
+//'         \item addv() [v] (add a vector v)
+//'         \item subv() [v] (subtract
+//'            a vector v);
+//'         }
+//'     \item vector reducing/augmenting function:\itemize{
+//'         \item diff() [k]
+//'         \item conv() [v, shape] (convolve with vector v; shape="full" is equivalent
+//'            to R's \code{convolve(..., rev(v), type="open")}).
+//'         \item quantile() [p] (calculate quantiles corresponding to probabilities p;
+//'            equivalent to R's \code{quantile(..., probs=p, type=8)}).
+//'        }
+//'     }
 //' 
 //' RcppArmadillo is used to do the job in very fast way but it comes at price
 //' of not allowing NA in the input numeric array.
@@ -83,17 +112,20 @@ case a_##act: \
 //' @param idim integer, dimension number along which a function must be applied
 //' @param fun character string, function name to be applied
 //' @param ... additional named parameters. Optional parameters can be helpful for
-//'    the following functions:
-//'       sd(), var() [norm_type: 0 normalisation using N-1 entries (default);
+//'    the following functions:\itemize{
+//'       \item sd(), var() [norm_type: 0 normalisation using N-1 entries (default);
 //'          1 normalisation using N entries];
-//'       norm() [p: integer >= 1 (default=2) or one of "-inf", "inf", "fro".]
-//'       normalise() [p: integer >= 1, default=2]
-//'       diff() [k: integer >= 1 (default=1) number of recursive application of diff().
+//'       \item norm() [p: integer >= 1 (default=2) or one of "-inf", "inf", "fro".]
+//'       \item normalise() [p: integer >= 1, default=2]
+//'       \item diff() [k: integer >= 1 (default=1) number of recursive application of diff().
 //'          The size of idim-th dimension will be reduced by k.]
-//'       trapz() [x: numerical vector of the same length as idim-th size of arr]
-//'    Mandatory parameter:
-//'       multv(), divv(), addv(), subv() [v: numerical vector of the same
+//'       \item trapz() [x: numerical vector of the same length as idim-th size of arr]
+//'    }
+//'    Mandatory parameter:\itemize{
+//'       \item multv(), divv(), addv(), subv() [v: numerical vector of the same
 //'          length as idim-th size of arr]
+//'       \item quantile() [p: vector of probabilities in interval [0; 1]]
+//'    }
 //'
 //' @return output array of dimension cut by 1 (the idim-th dimension
 //'    will disappear for reducing functions) or of the same dimension
@@ -145,15 +177,22 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
     add_map(subv);
     // vector reducing functions
     add_map(diff);
-   
+    add_map(conv);
+    add_map(quantile);
+  
+#define NBUF 512
+    char buf[NBUF];
+    if (mapf.count(fun) == 0) {
+        snprintf(buf, NBUF, "arrApply: fun='%s' is not in the list of applicable functions", fun.c_str());
+        stop(buf);
+    }
     array_act aact=mapf[fun];
-    std::vector<array_act> mvact={a_normalise, a_cumsum, a_cumprod, a_multv, a_divv, a_addv, a_subv, a_diff}; // mapping or vector reducing acts
+    std::vector<array_act> mvact={a_normalise, a_cumsum, a_cumprod, a_multv, a_divv, a_addv, a_subv, a_diff, a_conv, a_quantile}; // mapping or vector reducing acts
     bool mv_res=std::find(mvact.begin(), mvact.end(), aact) != mvact.end();
     uvec d;
-    char buf[512];
     // optional parameters
     unsigned int p=2, norm_type=0, k=1;
-    std::string pch="";
+    std::string pch="", shape="full";
     vec x;
     // mandatory parameter
     rowvec rowv;
@@ -161,11 +200,9 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
     bool p_is_int=true, use_x=false;
     RObject robj;
     double pr;
-    
-    if (mapf.count(fun) == 0) {
-        sprintf(buf, "arrApply: fun='%s' is not in the list of applicable functions", fun.c_str());
-        stop(buf);
-    }
+    //Rcout << "mapf.count('" << fun << "')=" <<  mapf.count(fun) << std::endl;
+    //for (auto it=mapf.cbegin(); it != mapf.cend(); ++it)
+    //    Rcout << it->first << "=>\t" << it->second << std::endl;
     // get args from dots
     switch(aact) {
         case a_sd:
@@ -174,7 +211,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                 //get "norm_type" in dots
                 norm_type=as<unsigned int>(dots["norm_type"]);
                 if (norm_type != 0 && norm_type != 1) {
-                    sprintf(buf, "arrApply: optional norm_type must be 0 or 1, instead got %d.", norm_type);
+                    snprintf(buf, NBUF, "arrApply: optional norm_type must be 0 or 1, instead got %d.", norm_type);
                     stop(buf);
                 }
             }
@@ -193,7 +230,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                         if (std::abs(round(pr)-pr) < 1.e-10) {
                             p=round(pr);
                         } else {
-                            sprintf(buf, "arrApply: optional p must be integer >= 1, instead got %g.", pr);
+                            snprintf(buf, NBUF, "arrApply: optional p must be integer >= 1, instead got %g.", pr);
                             stop(buf);
                         }
                     break;
@@ -202,15 +239,15 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                         p_is_int=false;
                     break;
                     default:
-                        sprintf(buf, "arrApply: unauthorized type for optional p (must be integer >= 1 or string), instead got SEXP type %d.", robj.sexp_type());
+                        snprintf(buf, NBUF, "arrApply: unauthorized type for optional p (must be integer >= 1 or string), instead got SEXP type %d.", robj.sexp_type());
                         stop(buf);
                     break;
                 }
                 if (!p_is_int && !(pch == "-inf" || pch == "inf" || pch == "fro")) {
-                    sprintf(buf, "arrApply: optional p, when a string, must be one of '-inf', 'inf' or 'fro'. Instead, got '%s'.", pch.c_str());
+                    snprintf(buf, NBUF, "arrApply: optional p, when a string, must be one of '-inf', 'inf' or 'fro'. Instead, got '%s'.", pch.c_str());
                     stop(buf);
                 } else if (p_is_int && p < 1) {
-                    sprintf(buf, "arrApply: optional p, when an integer, must be >= 1. Instead, got '%d'.", p);
+                    snprintf(buf, NBUF, "arrApply: optional p, when an integer, must be >= 1. Instead, got '%d'.", p);
                     stop(buf);
                 }
             }
@@ -220,7 +257,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                 //get "k" in dots
                 k=as<unsigned int>(dots["k"]);
                 if (k <= 0) {
-                    sprintf(buf, "arrApply: optional k must be an integer >= 1, instead got %d.", k);
+                    snprintf(buf, NBUF, "arrApply: optional k must be an integer >= 1, instead got %d.", k);
                     stop(buf);
                 }
             }
@@ -232,7 +269,8 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
             if (dots.containsElementNamed("v")) {
                 rowv=as<rowvec>(dots["v"]);
             } else {
-                sprintf(buf, "Parameter v is mandatory for %s() function", fun.c_str());
+                snprintf(buf, NBUF, "Parameter v is mandatory for %s() function", fun.c_str());
+                stop(buf);
             }
         break;
         case a_trapz:
@@ -241,6 +279,27 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                 use_x=true;
             }
         break;
+        case a_quantile:
+            if (dots.containsElementNamed("p")) {
+                x=as<vec>(dots["p"]);
+            } else {
+                snprintf(buf, NBUF, "Parameter p is mandatory for %s() function", fun.c_str());
+                stop(buf);
+            }
+        break;
+        case a_conv:
+            if (dots.containsElementNamed("v")) {
+                x=as<vec>(dots["v"]);
+            } else {
+                snprintf(buf, NBUF, "Parameter v is mandatory for %s() function", fun.c_str());
+                stop(buf);
+            }
+            if (dots.containsElementNamed("shape")) {
+                shape=as<std::string>(dots["shape"]);
+            } else {
+                shape=std::string("full");
+            }
+        break;
         default:
         ;
         break;
@@ -258,7 +317,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
         idim=1;
     }
     if (idim < 1 || idim > d.size()) {
-        sprintf(buf, "arrApply: idim (%d) is out of range [1, %d]", idim, d.size());
+        snprintf(buf, NBUF, "arrApply: idim (%d) is out of range [1, %d]", idim, d.size());
         stop(buf);
     }
     
@@ -287,6 +346,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
         dwork[1]=d[idim-1];
         dwork[2]=prod(d.tail(d.size()-idim));
     }
+    // dwork.print("dwork");
 //Rprintf("yah4\n");
     
     cube work(arr.begin(), dwork[0], dwork[1], dwork[2], false);
@@ -336,20 +396,43 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                 for (std::size_t isl=0; isl < work.n_slices; ++isl)
                     cres.slice(isl)=normalise(work.slice(isl), p, 1);
             } else {
-                sprintf(buf, "arrApply: for normalise() call, p is supposed to be integer, not a string.");
+                snprintf(buf, NBUF, "arrApply: for normalise() call, p is supposed to be integer, not a string.");
                 stop(buf);
             }
         break;
         case a_diff:
 //Rprintf("d1\n");
-              cres.set_size(dwork[0], dwork[1]-std::min(k, dwork[1]), dwork[2]);
+            cres.set_size(dwork[0], dwork[1]-std::min(k, dwork[1]), dwork[2]);
+//Rprintf("d2\n");
+//Rcout << "c.r=" << cres.n_rows << "; c.c=" << cres.n_cols << "; c.s=" << cres.n_slices << std::endl;
+            for (std::size_t isl=0; isl < work.n_slices; ++isl)
+                cres.slice(isl)=diff(work.slice(isl), k, 1);
+//Rprintf("d3\n");
+        break;
+        case a_quantile:
+            cres.set_size(dwork[0], x.size(), dwork[2]);
+            for (std::size_t isl=0; isl < work.n_slices; ++isl)
+                cres.slice(isl)=quantile(work.slice(isl), x, 1);
+        break;
+        case a_conv:
+//Rprintf("d1\n");
+            if (shape == "full")
+                cres.set_size(dwork[0], dwork[1]+x.n_elem-1, dwork[2]);
+            else if (shape == "same")
+                cres.set_size(dwork[0], dwork[1], dwork[2]);
+            else {
+                snprintf(buf, NBUF, "arrApply: for conv(), if given, the parameter shape must be 'full' (default) or 'same', got '%s'", shape.c_str());
+                stop(buf);
+            }
 //Rprintf("d2\n");
 //Rcout << "c.r=" << cres.n_rows << "; c.c=" << cres.n_cols << "; c.s=" << cres.n_slices << std::endl;
         {
-            mat mtmp;
+            rowvec mtmp;
             for (std::size_t isl=0; isl < work.n_slices; ++isl) {
-                mtmp=diff(work.slice(isl), k, 1);
-                cres.slice(isl)=mtmp;
+                for (std::size_t ir=0; ir < work.n_rows; ir++) {
+                    mtmp=conv(work.slice(isl).row(ir), x, shape.c_str());
+                    cres.slice(isl).row(ir)=mtmp;
+                }
             }
         }
 //Rprintf("d3\n");
@@ -381,7 +464,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                 break;
                 }
             } else {
-                sprintf(buf, "arrApply: for multv(), length(v) (%d) must be equal to dim(arr)[idim] (%d).", rowv.size(), work.n_cols);
+                snprintf(buf, NBUF, "arrApply: for multv() and alike, length(v) (%d) must be equal to dim(arr)[idim] (%d).", rowv.size(), work.n_cols);
                 stop(buf);
             }
         break;
@@ -391,7 +474,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
                     for (std::size_t isl=0; isl < work.n_slices; ++isl)
                         res(span::all,isl)=trapz(x, work.slice(isl), 1);
                 } else {
-                    sprintf(buf, "arrApply: for trapz(), length(x) (%d) must be equal to dim(arr)[idim] (%d).", x.size(), work.n_cols);
+                    snprintf(buf, NBUF, "arrApply: for trapz(), length(x) (%d) must be equal to dim(arr)[idim] (%d).", x.size(), work.n_cols);
                     stop(buf);
                 }
             } else {
@@ -400,7 +483,7 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
             }
         break;
         default:
-           stop("arrApply: It cannot be but unknown action is encountered.");
+           stop("arrApply: It cannot be, but unknown action is encountered.");
         break;
     }
     // rechape back the result
@@ -414,18 +497,18 @@ SEXP arrApply(NumericVector arr, unsigned int idim=1, std::string fun="sum", Lis
         v=lres;
         vres=wrap(v);
     } else if (mv_res) {
-        // mapping actions
+        // mapping/vec-reducing actions
         vres=wrap(cres);
+        if (d.size()) {
+            d[idim-1]=cres.n_cols;
+            //d.print("dim final");
+            vres.attr("dim")=IntegerVector(d.begin(), d.end());
+        } else {
+            vres.attr("dim")=R_NilValue;
+        }
     } else {
         // reducing actions
         vres=wrap(res);
     }
-    if (d.size()) {
-        if (aact == a_diff)
-            d[idim-1]-=std::min(k, d[idim-1]);
-        vres.attr("dim")=IntegerVector(d.begin(), d.end());
-    } else {
-        vres.attr("dim")=R_NilValue;
-    }
     return vres;
 }
